"""
ChecklistFabrik Markdown module

This module renders Jinja templated Markdown to HTML.

EXAMPLE::

    - linuxfabrik.clf.markdown:
        content: |
            ### Markdown Support

            ChecklistFabrik supports *Markdown*!
"""


def main(**kwargs):
    clf_jinja_env = kwargs['clf_jinja_env']
    clf_markdown = kwargs['clf_markdown']

    rendered_html = clf_markdown(
        clf_jinja_env.from_string(kwargs['content']).render(**kwargs),
    )

    return {
        'html': f'<div class="form-label">{rendered_html}</div>',
    }



"""
ChecklistFabrik select_input module

This module renders a simple HTML select (multi-selection as well as optgroup and hr elements are not supported).

EXAMPLE::

    - linuxfabrik.clf.select_input:
        label: 'Backup datacenter location for {{ host }}?'
        values:
          - 'Moon'
          - 'Mars'
          - 'Earth'
        required: true
        multiple: false
      fact_name: 'backup_datacenter_location'
"""

TEMPLATE_MULTI_SELECT_STRING = '''\
<div class="form-group">
    <div class="form-label" id="{{ fact_name }}-label">
        {% if required %}{% include "required_indicator.html.j2" %}{% endif %}
        <div>{% if not templated_label and required %}<i>A selection is required</i>{% endif %}{{ templated_label }}</div>
    </div>
    <select class="form-select" id="{{ fact_name }}" name="{{ fact_name }}[]" multiple
        aria-labelledby="{{ fact_name }}-label"
        {%- if required %} required{%- endif %}>
        {% for value in templated_values %}
        <option value="{{ value }}"{%- if value in fact_value %} selected{%- endif %}>{{ value }}</option>
        {% endfor %}
    </select>
    {# Hidden input to allow selecting no option, since an HTML form does not send empty selections. #}
    <input type="hidden" name="{{ fact_name }}[]" value="" />
</div>
'''

TEMPLATE_SINGLE_SELECT_STRING = '''\
<div class="form-group">
    <div class="form-label" id="{{ fact_name }}-label">
        {% if required %}{% include "required_indicator.html.j2" %}{% endif %}
        <div>{% if not templated_label and required %}<i>A selection is required</i>{% endif %}{{ templated_label }}</div>
    </div>
    <select class="form-select" id="{{ fact_name }}" name="{{ fact_name }}"
        aria-labelledby="{{ fact_name }}-label"
        {%- if required %} required{%- endif %}>
        <option value="">--- Please select ---</option>
        {% for value in templated_values %}
        <option value="{{ value }}"{%- if value == fact_value %} selected{%- endif %}>{{ value }}</option>
        {% endfor %}
    </select>
</div>
'''


def main(**kwargs):
    clf_jinja_env = kwargs['clf_jinja_env']
    clf_markdown = kwargs['clf_markdown']
    fact_name = kwargs['fact_name' if 'fact_name' in kwargs else 'auto_fact_name']

    templated_label = clf_markdown(clf_jinja_env.from_string(kwargs.get('label', '')).render(**kwargs))

    templated_values = [clf_jinja_env.from_string(value).render(**kwargs) for value in kwargs.get('values', [''])]

    if kwargs.get('multiple'):
        html = clf_jinja_env.from_string(
            TEMPLATE_MULTI_SELECT_STRING,
        ).render(
            **(kwargs | {
                'fact_name': fact_name,
                'fact_value': kwargs.get(fact_name, []),
                'templated_label': templated_label,
                'templated_values': templated_values,
            }),
        )
    else:
        html = clf_jinja_env.from_string(
            TEMPLATE_SINGLE_SELECT_STRING,
        ).render(
            **(kwargs | {
                'fact_name': fact_name,
                'fact_value': kwargs.get(fact_name),
                'templated_label': templated_label,
                'templated_values': templated_values,
            }),
        )

    return {
        'html': html,
        'fact_name': fact_name,
    }



"""
ChecklistFabrik checkbox_input module

This module renders either a single HTML checkbox input field or a group of them.

EXAMPLE::

    - linuxfabrik.clf.checkbox_input:
        label: 'Single checkbox input'
        required: false
      fact_name: 'single_check_result'

    - linuxfabrik.clf.checkbox_input:
        label: 'Use "values" if you want a group of checkboxes'
        values:
            - label: 'Step 1'
            - label: 'Step 2'
              value: 'step2'
            - value: 'Step 3'
        required: true
      fact_name: 'multi_check_result'
"""

import uuid

TEMPLATE_MULTI_CHECK_STRING = '''\
<fieldset>
    {% if templated_label %}
    <legend>
        {% if required %}{% include "required_indicator.html.j2" %}{% endif %}
        <span>{{ templated_label }}</span>
    </legend>
    {% endif %}

    {% for check in templated_checks %}
    <label class="form-checkbox">
        <input name="{{ fact_name }}[]" type="checkbox" value="{{ check.value }}"
            {%- if check.value in fact_value %} checked{%- endif %}
            {%- if check.required or required %} required{%- endif %} />
        <i class="form-icon"></i>
        <span>{% if check.required and not required %}{% include "required_indicator.html.j2" %}{% endif %}{{ check.templated_label | default(check.value, true) }}</span>
    </label>
    {% endfor %}

    {# Hidden field to allow unchecking all checkboxes, since an HTML form does not send unchecked checkboxes. #}
    <input type="hidden" name="{{ fact_name }}[]" value="" />
</fieldset>
'''

TEMPLATE_SINGLE_CHECK_STRING = '''\
<label class="form-checkbox">
    <input name="{{ fact_name }}" type="checkbox"
        {%- if fact_value %} checked{%- endif %}
        {%- if required %} required{%- endif %} />
    <i class="form-icon"></i>
    <span>{% if required %}{% include "required_indicator.html.j2" %}{% endif %}{% if not templated_label and required %}<i>An input is required</i>{% endif %}{{ templated_label }}</span>
</label>

{# Hidden field to allow unchecking a checkbox, since an HTML form does not send unchecked checkboxes. #}
<input type="hidden" name="{{ fact_name }}" value="" />
'''


def main(**kwargs):
    clf_jinja_env = kwargs['clf_jinja_env']
    clf_markdown= kwargs['clf_markdown']
    fact_name = kwargs['fact_name' if 'fact_name' in kwargs else 'auto_fact_name']

    templated_label = clf_markdown(clf_jinja_env.from_string(kwargs.get('label', '')).render(**kwargs))

    task_context_update = None

    if kwargs.get('values'):
        templated_checks = [
            {
                'label': check.get('label'),
                'templated_label': clf_markdown(clf_jinja_env.from_string(check['label']).render(**kwargs)) if check.get('label') else None,
                'value': check.get('value', uuid.uuid4().hex),
                'required': check.get('required'),
            }
            for check in kwargs['values']
        ]

        html = clf_jinja_env.from_string(
            TEMPLATE_MULTI_CHECK_STRING,
        ).render(
            **(kwargs | {
                'fact_name': fact_name,
                'fact_value': kwargs.get(fact_name, []),
                'templated_label': templated_label,
                'templated_checks': templated_checks,
            }),
        )

        task_context_update = {
            'values': [
                {
                    key: value
                    for key, value in check.items()
                    if key in ('label', 'value', 'required') and value is not None
                }
                for check in templated_checks
            ]
        }
    else:
        # If we don't have any values, just render a single checkbox.
        html = clf_jinja_env.from_string(
            TEMPLATE_SINGLE_CHECK_STRING,
        ).render(
            **(kwargs | {
                'fact_name': fact_name,
                'fact_value': kwargs.get(fact_name),
                'templated_label': templated_label,
            }),
        )

    return {
        'html': html,
        'fact_name': fact_name,
        'task_context_update': task_context_update,
    }



import mistune


class ClfHtmlRenderer(mistune.HTMLRenderer):
    def block_code(self, code, info=None):
        """Renders spectre.css compatible code blocks."""

        attributes = ''

        if info:
            attributes += f' data-lang="{mistune.util.escape(info.split(None, 1)[0])}"'

        lines = code.split('\n')
        if lines and lines[-1] == '':
            lines = lines[:-1]
        wrapped = ''.join(
            f'<span class="clf-code-line">{mistune.util.escape(line)}</span>'
            for line in lines
        )

        return (
            # DO NOT include line breaks / other whitespace in the `pre`-element as whitespace is displayed for `pre`-elements.
            f'<pre class="code"{attributes}>'
            '<button class="btn btn-link btn-sm clf-code-copy-btn" type="button">Copy</button>'
            f'<code>{wrapped}</code>'
            '</pre>'
        )

    def link(self, text, url, title=None):
        """Renders links as an HTML anchor with the `target="_blank"` attribute."""

        attributes = ''

        if title:
            attributes += f' title="{mistune.util.safe_entity(title)}"'

        return f'<a href="{self.safe_url(url)}" target="_blank"{attributes}>{text}</a>'


def create_markdown():
    return mistune.create_markdown(
        escape=False,
        renderer=ClfHtmlRenderer(),
        plugins=['strikethrough', 'table', 'speedup'],
    )



import importlib
import inspect
import logging
import uuid

import jinja2

from .. import utils

MODULE_NAMESPACE = 'checklistfabrik.modules'

TEMPLATE_FORMAT_STRING = '{html}\n<div class="divider"></div>'

logger = logging.getLogger(__name__)


class Task:
    """Models a ChecklistFabrik checklist task."""

    def __init__(self, module, context, fact_name, when, unnamed_fact=None):
        self.module = module
        self.context = context
        self.fact_name = fact_name
        self.when = when
        self.unnamed_fact = unnamed_fact

    def to_dict(self, facts):
        result = {
            self.module: self.context,
        }

        if self.fact_name is not None:
            result['fact_name'] = self.fact_name

            if facts is not None and self.fact_name in facts:
                result['value'] = facts[self.fact_name]

        if self.when is not None:
            result['when'] = self.when

        return result

    def eval_when(self, facts):
        """ Evaluate this page's 'when' condition(s) using provided facts."""

        try:
            result = utils.eval_when(facts, self.when)
        except jinja2.exceptions.TemplateSyntaxError as error:
            logger.error('Syntax error at "%s": %s', self.when, error.message)
            return False, f'<div class="toast toast-error">Syntax error at "{self.when}": {error.message}</div>'

        return result, None

    def render(self, facts, template_env, markdown):
        """Render the task using its module."""

        show_task, error = self.eval_when(facts)

        if not show_task:
            return ''
        elif error:
            return TEMPLATE_FORMAT_STRING.format(html=error)

        try:
            loaded_module = importlib.import_module(f'{MODULE_NAMESPACE}.{self.module}')
        except ModuleNotFoundError:
            logger.error('Task rendering error: Cannot find module "%s"', self.module)
            return f'<div class="toast toast-error">Task rendering error: Cannot find module <em>{self.module}</em>. Is it installed?</div>'

        render_context = facts.copy()
        render_context['clf_jinja_env'] = template_env
        render_context['clf_markdown'] = markdown
        if self.fact_name:
            render_context['fact_name'] = self.fact_name
        else:
            # Provide an automatic fact name that the module can use.
            # The module needs to report this back as fact_name if it uses the suggested name.
            auto_fact_name = f'auto_{uuid.uuid4().hex}'
            render_context['auto_fact_name'] = auto_fact_name  # Prefix the hex uuid so that it is a valid Python identifier.

            # Inject value that was provided without fact name.
            if self.unnamed_fact:
                render_context[auto_fact_name] = self.unnamed_fact
        render_context.update(self.context)

        if not hasattr(loaded_module, 'main') or not callable(loaded_module.main):
            logger.error(
                'Task rendering error: Module "%s" is not a valid ChecklistFabrik module as it has no callable "main"',
                self.module,
            )
            return f'<div class="toast toast-error">Task rendering error: Module <em>{self.module}</em> is not a valid ChecklistFabrik module.</div>'

        main_signature = inspect.signature(loaded_module.main)

        if not inspect.Parameter.VAR_KEYWORD in [param.kind for param in main_signature.parameters.values()]:
            logger.error(
                'Task rendering error: Module "%s" looks like a ChecklistFabrik module but is malformed as its signature does not allow variadic keyword arguments',
                self.module,
            )
            return TEMPLATE_FORMAT_STRING.format(
                html=f'<div class="toast toast-error">Task rendering error: Module <em>{self.module}</em> looks like a ChecklistFabrik module but is malformed.</div>',
            )

        try:
            result = loaded_module.main(**render_context)
        except Exception as exception:
            logger.error('Task rendering error: Rendering module "%s" failed: %s', self.module, exception)
            return TEMPLATE_FORMAT_STRING.format(
                html=f'<div class="toast toast-error">Task rendering error: Rendering module <em>{self.module}</em> failed: <pre>{exception}</pre></div>',
            )

        if not isinstance(result, dict):
            logger.error(
                'Task rendering error: Module "%s" returned a value of type "%s" but expected "%s"',
                self.module,
                type(result),
                type({}),
            )
            return TEMPLATE_FORMAT_STRING.format(
                html=f'<div class="toast toast-error">Task rendering error: Module <em>{self.module}</em> returned an invalid output.</div>',
            )

        if 'fact_name' in result:
            # Set fact name to the reported one from the module if we do not already have one.
            # Note that pure output modules (i.e. they do not render inputs) might not report a fact name (as they don't need it).
            if not self.fact_name:
                self.fact_name = result['fact_name']
            elif self.fact_name != result['fact_name']:
                logger.warning('Task module reports a different fact name than originally specified. This is most likely a bug in the task module')

        task_context_update = result.get('task_context_update')

        if task_context_update:
            self.context.update(task_context_update)

        return TEMPLATE_FORMAT_STRING.format(html=result.get('html', ''))



import logging
import webbrowser
import wsgiref.simple_server

logger = logging.getLogger(__name__)


class ChecklistWsgiRequestHandler(wsgiref.simple_server.WSGIRequestHandler):
    """A variant of the WSGIRequestHandler that logs messages using the logging package."""

    def log_message(self, format, *args):
        if not 'GET /heartbeat HTTP/1.1' in args:
            logger.debug(format, *args)


class ChecklistWsgiServer(wsgiref.simple_server.WSGIServer):
    """The WSGI server that powers the ChecklistFabrik HTML interface."""

    def __init__(self, host, port, application):
        self.exit_flag = False

        super().__init__((host, port), ChecklistWsgiRequestHandler)

        self.set_app(application)

        application.server_exit_callback = self.exit

    def exit(self):
        self.exit_flag = True

    def serve(self, open_browser=False):
        host, port = self.server_address

        logger.info('Starting server on "http://%s:%d"', host, port)

        print('Press Ctrl+C to shutdown the server and save the checklist to disk.')

        if open_browser:
            webbrowser.open(f'http://{host}:{port}')

        try:
            while not self.exit_flag:
                self.handle_request()
        except KeyboardInterrupt:
            pass
        finally:
            logger.info('Shutting down server')
            self.server_close()



import argparse
import logging
import pathlib

import ruamel.yaml

from .base_cli import BaseCli, IntRange
from .. import __version__
from .. import checklist_data_mapper
from .. import checklist_wsgi_app
from .. import checklist_wsgi_server
from .. import dashboard_wsgi_app
from .. import templates

DESCRIPTION = (
    'Interactive CLI for launching dynamic, web-based checklists. '
    'Leverage YAML templates with Jinja logic to create, run, and track recurring procedures. '
    'Run without arguments to open the dashboard.'
)
HOST = '127.0.0.1'

logger = logging.getLogger(__name__)

__author__ = 'Linuxfabrik GmbH, Zurich/Switzerland'


class PlayCli(BaseCli):
    """The ChecklistFabrik play CLI."""

    def __init__(self):
        super().__init__(DESCRIPTION)

        self.yaml = ruamel.yaml.YAML()

        self.yaml.preserve_quotes = True
        self.yaml.block_seq_indent = 2
        self.yaml.map_indent = 2
        self.yaml.sequence_indent = 4

        self.data_mapper = checklist_data_mapper.ChecklistDataMapper(self.yaml)

    def init_args(self):
        self.arg_parser.add_argument(
            '-V', '--version',
            help='Display the program\'s version information and exit.',
            action='version',
            version=f'%(prog)s: v{__version__} by {__author__}'
        )

        self.arg_parser.add_argument(
            '-v', '--verbose',
            action='store_true',
            help='Optional: Also log debug messages on console.',
        )

        self.arg_parser.add_argument(
            'report_file',
            help=(
                'Path to the report file. If the file exists, it will be loaded for re-running. '
                'If you want to create a new report from an existing checklist/template, '
                'provide a non-existent file path and use the `--template` option. '
                'This option may be left empty to auto-generate the file path '
                'based on the template\'s `report_path` (if provided) or simply a timestamp. '
                'If omitted entirely (along with `--template`), the dashboard is opened.'
            ),
            nargs='?',
            type=pathlib.Path,
        )

        self.arg_parser.add_argument(
            '--force',
            action='store_true',
            help=(
                'Allow creating a checklist from a template even if the report checklist file '
                '(the `report_file` argument) already exists. '
                'WARNING: THE REPORT FILE WILL BE OVERWRITTEN.'
            ),
        )

        self.arg_parser.add_argument(
            '--open',
            action=argparse.BooleanOptionalAction,
            help='Control whether to open the checklist page the default browser.',
            default=True,
        )

        self.arg_parser.add_argument(
            '--port',
            help='Port to use for the HTTP server. Using "0" will auto-select an available port. Default: %(default)d',
            default=0,
            type=IntRange(0, 65535),
        )

        self.arg_parser.add_argument(
            '--reports-dir',
            help='Directory to scan for report files (dashboard mode). Default: %(default)s',
            type=pathlib.Path,
            default=pathlib.Path('.'),
        )

        self.arg_parser.add_argument(
            '--template',
            help=(
                'Optional: Path to a YAML template file for creating a new report. '
                'This option may only be used when the report file (the `report_file` '
                'argument) does not already exist or the `--force` option is used.'
            ),
            type=pathlib.Path,
        )

        self.arg_parser.add_argument(
            '--templates-dir',
            help='Directory to scan for checklist templates (dashboard mode). Default: %(default)s',
            type=pathlib.Path,
            default=pathlib.Path('.'),
        )

    def validate_args(self):
        if self.args.report_file is None and self.args.template is None:
            # Dashboard mode.
            # Auto-detect well-known subdirectories if the user did not override.
            if self.args.reports_dir == pathlib.Path('.') and pathlib.Path('reports').is_dir():
                self.args.reports_dir = pathlib.Path('reports')
            if self.args.templates_dir == pathlib.Path('.') and pathlib.Path('templates').is_dir():
                self.args.templates_dir = pathlib.Path('templates')
            return

        if self.args.template is not None:
            if self.args.report_file and self.args.report_file.is_file() and not self.args.force:
                self.arg_parser.error('--template may only be specified if the report file does not exist')

            if not self.args.template.is_file():
                self.arg_parser.error('--template must be a file')

        else:
            if not (self.args.report_file and self.args.report_file.is_file()):
                self.arg_parser.error('report file must exist')

    def run(self):
        if self.args.report_file is None and self.args.template is None:
            return self.run_dashboard()

        checklist_app = checklist_wsgi_app.ChecklistWsgiApp(
            self.args.report_file,
            self.data_mapper,
            templates.get_template_loader(),
            templates.get_assets_path(),
            checklist_template=self.args.template,
        )

        checklist_server = checklist_wsgi_server.ChecklistWsgiServer(HOST, self.args.port, checklist_app)

        checklist_server.serve(open_browser=self.args.open)

        checklist_app.save_checklist()

        return 0

    def run_dashboard(self):
        dashboard_app = dashboard_wsgi_app.DashboardWsgiApp(
            self.args.templates_dir.resolve(),
            self.args.reports_dir.resolve(),
            self.data_mapper,
            templates.get_template_loader(),
            templates.get_assets_path(),
        )

        dashboard_server = checklist_wsgi_server.ChecklistWsgiServer(HOST, self.args.port, dashboard_app)

        dashboard_server.serve(open_browser=self.args.open)

        dashboard_app.cleanup()

        return 0


def main(args=None):
    PlayCli.main(args)



import argparse
import importlib.metadata


def main():
    distribution = importlib.metadata.distribution('checklistfabrik')

    entry_points = {
        entry_point.name: entry_point
        for entry_point in distribution.entry_points
        if entry_point.group == 'console_scripts'
    }

    parser = argparse.ArgumentParser(prog='python -m checklistfabrik.core', add_help=False)
    parser.add_argument('entry_point', choices=entry_points.keys())
    args, extra = parser.parse_known_args()

    entry_points[args.entry_point].load()([args.entry_point] + extra)


if __name__ == '__main__':
    main()



"""
ChecklistFabrik import module stub

This module is a stub as it represents core functionality.

EXAMPLE::

    - linuxfabrik.clf.import: 'file-to-import.yml'
"""

HTML = '''\
<div class="toast toast-error">
    <em>linuxfabrik.clf.import</em> represents the core import functionality and should never actually be rendered on a checklist.<br />
    If you encounter this message on your checklist, it indicates that an unreachable state has been reached. Please report this issue.
</div>
'''


def main(**kwargs):
    return {
        'html': HTML,
    }



"""Package that contains the ChecklistFabrik Jinja templates and assets."""

import pathlib

import jinja2


def get_assets_path():
    return pathlib.Path(__file__).parent / 'assets'


def get_template_loader():
    return jinja2.PackageLoader(__name__, '.')


def get_template_path():
    return pathlib.Path(__file__).parent



import logging

import jinja2.exceptions

from .. import utils

TEMPLATE_FORMAT_STRING = '''\
<fieldset>
    <legend>{title}</legend>
    {data}
</fieldset>
'''

logger = logging.getLogger(__name__)


def remove_last_divider(html: str) -> str:
    """
    HTML hack: Remove the last horizontal line (if any) in a fieldset for aesthetic reasons.
    """
    divider = '<div class="divider"></div>'
    last_index = html.rfind(divider)
    if last_index != -1:
        return html[:last_index] + html[last_index + len(divider):]
    return html


class Page:
    """Models a ChecklistFabrik checklist page."""

    def __init__(self, title, tasks, when):
        self.title = title
        self.tasks = tasks
        self.when = when

    def to_dict(self, facts):
        result = {
            'title': self.title,
            'tasks': [task.to_dict(facts) for task in self.tasks],
        }

        if self.when is not None:
            result['when'] = self.when

        return result

    def eval_when(self, facts):
        """ Evaluate this task's "when" condition(s) using provided facts."""

        try:
            result = utils.eval_when(facts, self.when)
        except jinja2.exceptions.TemplateSyntaxError as error:
            logger.error('Syntax error at "%s": %s', self.when, error.message)
            return False, f'<div class="toast toast-error">Syntax error at "{self.when}": {error.message}</div>'

        return result, None

    def render(self, facts, template_env, markdown):
        """Render the page with all tasks using Jinja."""

        show_page, error = self.eval_when(facts)

        if show_page:
            data = ''.join([task.render(facts, template_env, markdown) for task in self.tasks])
            data = remove_last_divider(data)
        elif error:
            data = error
        else:
            data = '<div class="toast toast-primary">This page was marked as not applicable based on previous input.</div>'

        return TEMPLATE_FORMAT_STRING.format(
            title=template_env.from_string(self.title).render(**facts),
            data=data,
        )



"""General purpose functionalities for ChecklistFabrik."""

import jinja2


def eval_conditional(facts, conditional):
    """Evaluates a single jinja conditional statement."""

    # Uses spaces around True/False to ensure the result is still a string in case of rendering with jinja's NativeEnvironment.
    condition_template = jinja2.Template(f'{{% if {conditional} %}} True {{% else %}} False {{% endif %}}')

    return condition_template.render(**facts).strip() == 'True'


def eval_all_conditionals(facts, conditionals):
    """Conjunctive evaluation of a list of jinja conditional statements."""

    for conditional in conditionals:
        if not eval_conditional(facts, conditional):
            return False

    return True


def eval_when(facts, when):
    """
    Automagically evaluate either a when condition or list of conditions.

    The absence of "when" conditions is considered to be truthy.
    """

    if when is None:
        return True, None

    single_condition = isinstance(when, str) and eval_conditional(facts, when)
    multi_conditions = isinstance(when, list) and eval_all_conditionals(facts, when)

    return single_condition or multi_conditions


def validate_dict_keys(
        dictionary,
        required_keys,
        optional_keys=None,
        disallow_extra_keys=False,
):
    """Validate that a dictionary contains required keys and optional keys; if desired, also checks if any extra keys are present."""

    if required_keys is None:
        required_keys = set()

    if optional_keys is None:
        optional_keys = set()

    missing_keys = required_keys - dictionary.keys()

    if len(missing_keys) > 0:
        return False, f'Missing the following required keys: {", ".join(missing_keys)}'

    if disallow_extra_keys:
        extra_keys = dictionary.keys() - required_keys - optional_keys

        if len(extra_keys) > 0:
            return False, f'Unexpected extra keys: "{", ".join(extra_keys)}"'

    return True, 'Valid'



"""
ChecklistFabrik radio_input module

This module renders an HTML radio group.

EXAMPLE::

    - linuxfabrik.clf.radio_input:
        label: 'Choose your favourite food'
        values:
          - label: 'Pizza'
            value: 'pizza'
          - value: 'burger'
          - label: 'Other'
        required: true
      fact_name: 'fav_food'
"""

import uuid

TEMPLATE_STRING = '''
<fieldset>
    <legend>
        {% if required %}{% include "required_indicator.html.j2" %}{% endif %}
        <span>{% if not templated_group_label and required %}<i>A selection is required</i>{% endif %}{{ templated_group_label }}</span>
    </legend>

    {% for radio in templated_radios %}
    <label class="form-radio">
        <input name="{{ fact_name }}" type="radio" value="{{ radio.value }}"
            {%- if radio.value == fact_value %} checked{%- endif %}
            {%- if required %} required{%- endif %} />
        <i class="form-icon"></i>
        <span>{{ radio.templated_label | default(radio.value, true) }}</span>
    </label>
    {% endfor %}
</fieldset>
'''


def main(**kwargs):
    clf_jinja_env = kwargs['clf_jinja_env']
    clf_markdown = kwargs['clf_markdown']
    fact_name = kwargs['fact_name' if 'fact_name' in kwargs else 'auto_fact_name']

    templated_group_label = clf_markdown(clf_jinja_env.from_string(kwargs.get('label', '')).render(**kwargs))

    templated_radios = [
        {
            'label': radio.get('label'),
            'templated_label': clf_markdown(
                clf_jinja_env.from_string(radio['label']).render(**kwargs),
            ) if radio.get('label') else None,
            'value': radio.get('value', uuid.uuid4().hex),
        }
        for radio in kwargs.get('values', [])
    ]

    return {
        'html': clf_jinja_env.from_string(
            TEMPLATE_STRING,
        ).render(
            **(kwargs | {
                'fact_name': fact_name,
                'fact_value': kwargs.get(fact_name),
                'templated_group_label': templated_group_label,
                'templated_radios': templated_radios,
            }),
        ),
        'fact_name': fact_name,
        'task_context_update': {
            'values': [
                {
                    key: value
                    for key, value in radio.items()
                    if key in ('label', 'value') and value is not None
                }
                for radio in templated_radios
            ]
        },
    }



"""
ChecklistFabrik text_input module

This module renders an HTML text input field.

EXAMPLE:
    - linuxfabrik.clf.text_input:
        label: 'How many backups to keep for host {{ host }}?'
        required: true
      fact_name: 'nr_backups'
"""

TEMPLATE_STRING = '''
<div class="form-group">
    <div class="form-label" id="{{ fact_name }}-label">
        {% if required %}{% include "required_indicator.html.j2" %}{% endif %}
        <div>{% if not templated_label and required %}<i>An input is required</i>{% endif %}{{ templated_label }}</div>
    </div>
    <input class="form-input" id="{{ fact_name }}" name="{{ fact_name }}" type="text"
        aria-labelledby="{{ fact_name }}-label"
        {%- if required %} required{%- endif %}
        {%- if fact_value %} value="{{ fact_value }}"{%- endif %} />
</div>
'''


def main(**kwargs):
    clf_jinja_env = kwargs['clf_jinja_env']
    clf_markdown = kwargs['clf_markdown']
    fact_name = kwargs['fact_name' if 'fact_name' in kwargs else 'auto_fact_name']

    templated_label = clf_markdown(clf_jinja_env.from_string(kwargs.get('label', '')).render(**kwargs))

    return {
        'html': clf_jinja_env.from_string(
            TEMPLATE_STRING,
        ).render(
            **(kwargs | {
                'fact_name': fact_name,
                'fact_value': kwargs.get(fact_name),
                'templated_label': templated_label,
            }),
        ),
        'fact_name': fact_name,
    }



import json
import logging
import pathlib
import threading

import jinja2
import werkzeug
import werkzeug.exceptions
import werkzeug.middleware.shared_data
import werkzeug.routing
import werkzeug.utils

from . import checklist_wsgi_app
from . import checklist_wsgi_server

HOST = '127.0.0.1'

logger = logging.getLogger(__name__)


class DashboardWsgiApp:
    """The WSGI app that powers the ChecklistFabrik dashboard."""

    def __init__(self, templates_dir, reports_dir, data_mapper, template_loader, assets_dir):
        self.assets_dir = assets_dir
        self.data_mapper = data_mapper
        self.reports_dir = reports_dir
        self.server_exit_callback = None
        self.spawned_checklists = []
        self.template_loader = template_loader
        self.templates_dir = templates_dir

        self.templ_env = jinja2.Environment(loader=template_loader)

        self.url_map = werkzeug.routing.Map(
            [
                werkzeug.routing.Rule('/', endpoint=self.on_dashboard),
                werkzeug.routing.Rule('/run', endpoint=self.on_run, methods=['POST']),
                werkzeug.routing.Rule('/view', endpoint=self.on_view, methods=['POST']),
            ],
        )

        self.wsgi_app = werkzeug.middleware.shared_data.SharedDataMiddleware(
            self.application,
            {
                '/assets': str(assets_dir),
            },
        )

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    @werkzeug.Request.application
    def application(self, request):
        urls = self.url_map.bind_to_environ(request.environ)

        try:
            endpoint, values = urls.match()
            return endpoint(request, **values)
        except werkzeug.exceptions.HTTPException as exception:
            return exception

    def cleanup(self):
        """Shut down all spawned checklist servers and save their data."""
        for _app, server, thread in self.spawned_checklists:
            if thread.is_alive():
                server.exit()
                thread.join(timeout=10)

    def scan_directory(self, directory):
        """Scan a directory for checklist YAML files and extract metadata."""
        items = []

        if not directory.is_dir():
            return items

        for path in sorted(directory.glob('*.yml')):
            try:
                data = self.data_mapper.load_yaml(path)

                if not isinstance(data, dict) or 'title' not in data or 'pages' not in data:
                    continue

                items.append({
                    'description': str(data.get('description', '')),
                    'filename': path.name,
                    'path': str(path.resolve()),
                    'title': str(data.get('title', '')),
                })
            except Exception:
                continue

        return items

    def on_dashboard(self, request, **kwargs):
        if self.templates_dir.resolve() == self.reports_dir.resolve():
            # Same directory for both: show all files in both sections.
            # The user decides whether to "Run" (new from template) or "View" (re-open).
            all_items = self.scan_directory(self.templates_dir)
            templates_list = all_items
            reports_list = all_items
        else:
            templates_list = self.scan_directory(self.templates_dir)
            reports_list = self.scan_directory(self.reports_dir)

        return werkzeug.Response(
            self.templ_env.get_template('dashboard.html.j2').render(
                reports=reports_list,
                templates=templates_list,
            ),
            mimetype='text/html',
        )

    def _launch_checklist(self, checklist_file, checklist_template, allowed_dir):
        """Launch a new checklist server and return the URL as a JSON response."""
        path = checklist_file or checklist_template

        # Security: ensure path is within allowed directory.
        if not path.resolve().is_relative_to(allowed_dir.resolve()):
            raise werkzeug.exceptions.Forbidden()

        if not path.is_file():
            raise werkzeug.exceptions.NotFound()

        try:
            app = checklist_wsgi_app.ChecklistWsgiApp(
                checklist_file,
                self.data_mapper,
                self.template_loader,
                self.assets_dir,
                checklist_template=checklist_template,
            )
        except SystemExit:
            return werkzeug.Response(
                json.dumps({'error': 'Failed to load checklist'}),
                status=500,
                mimetype='application/json',
            )

        server = checklist_wsgi_server.ChecklistWsgiServer(HOST, 0, app)
        host, port = server.server_address

        def serve_and_save():
            server.serve(open_browser=False)
            app.save_checklist()

        thread = threading.Thread(target=serve_and_save, daemon=True)
        thread.start()

        self.spawned_checklists.append((app, server, thread))

        return werkzeug.Response(
            json.dumps({'url': f'http://{host}:{port}'}),
            mimetype='application/json',
        )

    def _parse_path_from_request(self, request):
        """Extract and validate the path from a JSON request body."""
        try:
            data = json.loads(request.data)
        except json.JSONDecodeError:
            raise werkzeug.exceptions.BadRequest()

        path = data.get('path')

        if not path:
            raise werkzeug.exceptions.BadRequest()

        return path

    def on_run(self, request, **kwargs):
        """Start a new checklist from a template."""
        template_path = self._parse_path_from_request(request)

        return self._launch_checklist(
            checklist_file=None,
            checklist_template=pathlib.Path(template_path),
            allowed_dir=self.templates_dir,
        )

    def on_view(self, request, **kwargs):
        """Open an existing report for viewing or re-playing."""
        report_path = self._parse_path_from_request(request)

        return self._launch_checklist(
            checklist_file=pathlib.Path(report_path),
            checklist_template=None,
            allowed_dir=self.reports_dir,
        )



import datetime
import json
import logging
import os.path
import pathlib
import uuid

import jinja2
import werkzeug
import werkzeug.exceptions
import werkzeug.middleware.shared_data
import werkzeug.routing
import werkzeug.utils

from . import markdown

TEMPLATE_STRING = '''\
{% extends "checklist.html.j2" %}
{% block form_content %}
{{data}}
{% endblock %}
'''

logger = logging.getLogger(__name__)

# Global Jinja variables for use in templates
jinja_globals = {
    'now': datetime.datetime.now,
}


class ChecklistWsgiApp:
    """The WSGI app that powers the ChecklistFabrik HTML interface."""

    def __init__(self, checklist_file, checklist_mapper, template_loader, assets_dir, checklist_template=None):
        self.checklist_file = checklist_file
        self.checklist_mapper = checklist_mapper
        self.checklist_template = checklist_template
        self.server_exit_callback = None

        self.server_id = uuid.uuid4().hex
        self.templ_env = jinja2.Environment(loader=template_loader)
        self.markdown = markdown.create_markdown()

        self.templ_env.globals.update(jinja_globals)

        self.url_map = werkzeug.routing.Map(
            [
                werkzeug.routing.Rule('/', endpoint=lambda request: werkzeug.utils.redirect('/page/0')),
                werkzeug.routing.Rule('/done', endpoint=self.on_done),
                werkzeug.routing.Rule('/exit', endpoint=self.on_exit),
                werkzeug.routing.Rule('/heartbeat', endpoint=self.on_heartbeat),
                werkzeug.routing.Rule('/page/', endpoint=lambda request: werkzeug.utils.redirect('/page/0')),
                werkzeug.routing.Rule('/page/<int:id>', endpoint=self.on_page_get, methods=['GET']),
                werkzeug.routing.Rule('/page/<int:id>', endpoint=self.on_page_post, methods=['POST']),
                werkzeug.routing.Rule('/page/<int:id>/next', endpoint=self.on_next_page),
                werkzeug.routing.Rule('/page/<int:id>/prev', endpoint=self.on_prev_page),
            ],
        )

        self.wsgi_app = werkzeug.middleware.shared_data.SharedDataMiddleware(
            self.application,
            {
                '/assets': str(assets_dir),
            },
        )

        self.checklist = self.load_checklist()

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def load_checklist(self):
        file_to_load = self.checklist_template if self.checklist_template is not None else self.checklist_file

        return self.checklist_mapper.load_checklist(file_to_load)

    def save_checklist(self):
        if self.checklist_file:
            # Loaded from this file, now write to it again.
            self.checklist_mapper.save_checklist(self.checklist_file, self.checklist)
            return

        # No file was specified, neither on the CLI nor in the template file, so generate one.

        # Try to generate a filename based on the template's report path.
        if self.checklist.report_path:
            generated_filename = self.templ_env.from_string(self.checklist.report_path).render(self.checklist.facts)

            # Remove well-known invalid characters for the most commonly used operating systems and filesystems.
            clean_filename = generated_filename.translate(
                str.maketrans(
                    {
                        ctrl_char: None for ctrl_char in range(0, 32)
                    } | {
                        '"': None,
                        '*': None,
                        ':': None,
                        '<': None,
                        '>': None,
                        '?': None,
                    }
                )
            )

            file_to_save = pathlib.Path(os.path.expandvars(clean_filename))
            logger.info('Generated file path based on template: "%s"', file_to_save.resolve())
        else:
            file_to_save = pathlib.Path(f'checklist_{datetime.date.today().isoformat()}.yml')
            logger.info('No report path configured. Using "%s"', file_to_save.resolve())

        # Saving to the file with the generated file name.
        # This should never overwrite already existing files with the same name.
        self.checklist_mapper.save_checklist(file_to_save, self.checklist, overwrite=False)

    @werkzeug.Request.application
    def application(self, request):
        urls = self.url_map.bind_to_environ(request.environ)

        try:
            endpoint, values = urls.match()
            return endpoint(request, **values)
        except werkzeug.exceptions.HTTPException as exception:
            return exception

    def on_page_get(self, request, **kwargs):
        page_id = kwargs['id']

        if page_id >= len(self.checklist):  # page_id will always be an unsigned integer due to Werkzeug's IntegerConverter.
            raise werkzeug.exceptions.NotFound()

        page_data = self.checklist.pages[page_id].render(self.checklist.facts, self.templ_env, self.markdown)

        return werkzeug.Response(
            self.templ_env.from_string(TEMPLATE_STRING).render(
                title=self.checklist.title,
                version=self.checklist.version,
                page_id=page_id,
                data=page_data,
                pages=[
                    {
                        'id': id,
                        'title': page.title,
                        'eval_when': page.eval_when(self.checklist.facts)[0],
                    }
                    for id, page in enumerate(self.checklist.pages)
                ],
                server_id=self.server_id,
            ),
            mimetype='text/html',
        )

    def on_page_post(self, request, **kwargs):
        page_id = kwargs['id']

        if page_id >= len(self.checklist):  # page_id will always be an unsigned integer due to Werkzeug's IntegerConverter.
            raise werkzeug.exceptions.NotFound()

        redirect = ''

        for key in request.form.keys():
            if key == 'submit_action':
                redirect = request.form.get(key, '').lower()
                continue

            if key.endswith('[]'):
                # List keys are marked with '[]' to differentiate them from single value keys,
                # otherwise it would be impossible to differentiate single values from lists with exactly one value (due to how HTML forms work).
                self.checklist.facts[key[:-2]] = [
                    value
                    for value in request.form.getlist(key)
                    # Only save if non-empty. An empty string is used to force the HTML form to send empty selections/states.
                    # This was implemented as otherwise it would be impossible to change an already submitted value
                    # to be blank (e.g. unchecking a checkbox) as the HTML form does not send empty inputs.
                    if value
                ]
            else:
                value = request.form.get(key)

                # Only save if non-empty, otherwise reset. An empty string is used to force the HTML form to send
                # empty selections/states. This was implemented as otherwise it would be impossible to change an already
                # submitted value to be blank (e.g. unchecking a checkbox) as the HTML form does not send empty inputs.
                self.checklist.facts[key] = value if value else None

        if redirect.startswith('page '):
            return werkzeug.utils.redirect(f'/page/{redirect.split(" ", maxsplit=1)[1]}')

        if redirect == 'next':
            return werkzeug.utils.redirect(f'/page/{page_id}/next')

        if redirect == 'previous':
            return werkzeug.utils.redirect(f'/page/{page_id}/prev')

        if redirect == 'save and exit':
            return werkzeug.utils.redirect('/exit')

        return werkzeug.Response(response='OK', status=200)

    def on_next_page(self, request, **kwargs):
        # Find the next applicable page.
        next_page_id = kwargs['id'] + 1
        while next_page_id < len(self.checklist):
            if self.checklist.pages[next_page_id].eval_when(self.checklist.facts)[0]:
                break

            next_page_id += 1
        else:
            return werkzeug.utils.redirect(f'/done')

        return werkzeug.utils.redirect(f'/page/{next_page_id}')

    def on_prev_page(self, request, **kwargs):
        # Find the last previously applicable page.
        prev_page_id = kwargs['id'] - 1
        while prev_page_id >= 0:
            if self.checklist.pages[prev_page_id].eval_when(self.checklist.facts)[0]:
                break

            prev_page_id -= 1
        else:
            # All previous pages are hidden; stay on the first page.
            prev_page_id = 0

        return werkzeug.utils.redirect(f'/page/{prev_page_id}')

    def on_done(self, request, **kwargs):
        return werkzeug.Response(
            self.templ_env.get_template('done.html.j2').render(
                last_page_id=len(self.checklist) - 1,
            ),
            mimetype='text/html',
        )

    def on_exit(self, request, **kwargs):
        if self.server_exit_callback is None or not callable(self.server_exit_callback):
            raise werkzeug.exceptions.NotImplemented()

        self.server_exit_callback()

        return werkzeug.Response(self.templ_env.get_template('shutdown.html.j2').render(), mimetype='text/html')

    def on_heartbeat(self, request, **kwargs):
        return werkzeug.Response(json.dumps({'server_id': self.server_id}), mimetype='application/json')



import argparse
import logging
import os
import pathlib
import platform
import sys


class BaseCli:
    """A base CLI implementation of common code for building complete CLIs."""

    def __init__(self, description=None):
        self.arg_parser = argparse.ArgumentParser(description=description)
        self.args = None
        self.logger = None

    def init_args(self):
        raise NotImplementedError('This method must be implemented by a subclass')

    def parse_args(self, args):
        self.arg_parser.prog = os.path.basename(args[0])
        self.args = self.arg_parser.parse_args(args[1:])

    def validate_args(self):
        pass

    def run(self):
        raise NotImplementedError('This method must be implemented by a subclass')

    def init_logging(self, console_log_level=logging.INFO, file_log_level=logging.DEBUG):
        root_module_name = __name__.split('.', maxsplit=1)[0]
        self.logger = logging.getLogger(root_module_name)
        self.logger.setLevel(min(console_log_level, file_log_level))

        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(console_log_level)
        console_handler.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))

        self.logger.addHandler(console_handler)

        instance_module_name = type(self).__module__.rsplit(".", maxsplit=1)[-1]
        log_file_name = f'{root_module_name}-{instance_module_name}.log'
        os_name = platform.system()

        if os_name == 'Darwin':
            log_path = pathlib.Path.home() / 'Library' / 'Logs' / root_module_name / log_file_name
        elif os_name == 'Linux':
            log_path = pathlib.Path(os.getenv('XDG_DATA_HOME', pathlib.Path.home() / '.local' / 'share')) / log_file_name
        elif os_name == 'Windows' and 'APPDATA' in os.environ:
            log_path = pathlib.Path(os.environ['APPDATA']) / root_module_name / log_file_name
        else:
            log_path = pathlib.Path.cwd() / log_file_name

        try:
            log_path.parent.mkdir(parents=True, exist_ok=True)
            log_path.touch(mode=0o644, exist_ok=True)
        except FileExistsError:
            self.logger.error('Cannot create/open log file in "%s" as it is not a directory', log_path.parent)
        except PermissionError:
            self.logger.error('Cannot write to log file "%s"', log_path)
        else:
            self.logger.info('Writing log to file "%s"', log_path)

            file_handler = logging.FileHandler(log_path, mode='w')
            file_handler.setLevel(file_log_level)
            file_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))

            self.logger.addHandler(file_handler)

    @classmethod
    def main(cls, args=None):
        if args is None:
            args = sys.argv

        cli = cls()

        cli.init_args()
        cli.parse_args(args)
        cli.validate_args()

        cli.init_logging(
            console_log_level=logging.DEBUG if getattr(cli.args, 'verbose', False) else logging.INFO,
        )

        sys.exit(cli.run())


class IntRange:
    """
    "Type" for argparse to check if an integer is in a given range.

    Use this instead of `choices=range(min, max + 1)` to avoid printing each
    possible value on the help text (especially helpful for large ranges).
    """

    def __init__(self, min, max):
        self.min = min
        self.max = max

    def __call__(self, string):
        value = None

        try:
            value = int(string)

            if not (self.min <= value <= self.max):
                raise ValueError

            return value
        except ValueError:
            msg = f'must be an integer in range [{self.min}, {self.max}]'

            if type(value) == int:
                msg += f', got {value}'

            raise argparse.ArgumentTypeError(msg)



"""
ChecklistFabrik HTML module

This module simply renders Jinja templated HTML.

EXAMPLE::

    - linuxfabrik.clf.html:
        content: 'This is an example text with Jinja expressions, for example {{ host }}.'
"""


def main(**kwargs):
    clf_jinja_env = kwargs['clf_jinja_env']

    rendered_content = clf_jinja_env.from_string(kwargs['content']).render(**kwargs)

    return {
        'html': f'<div class="clf-content-block">{rendered_content}</div>',
    }



class Checklist:
    """Models a ChecklistFabrik checklist."""

    def __init__(self, title, pages, facts, description=None, report_path=None, version=None):
        self.description = description
        self.facts = facts
        self.pages = pages
        self.report_path = report_path
        self.title = title
        self.version = version

    def __len__(self):
        return len(self.pages)

    def to_dict(self):
        result = {
            'title': self.title,
            'pages': [page.to_dict(self.facts) for page in self.pages],
        }

        if self.version is not None:
            result['version'] = self.version

        return result



import io
import logging
import pathlib
import sys

from . import models
from . import utils

logger = logging.getLogger(__name__)


class ChecklistLoadError(Exception):
    """Failure while loading checklist data."""
    pass


class ChecklistDataMapper:
    """Helper to map checklist data from YAML files to Python classes and vice versa."""

    def __init__(self, yaml):
        self.yaml = yaml

    def load_yaml(self, file):
        try:
            with open(file, mode='r', encoding='utf-8') as file_handle:
                return self.yaml.load(file_handle.read())
        except FileNotFoundError:
            logger.critical('Cannot open file "%s" as it does not exist', file)
            raise ChecklistLoadError
        except IsADirectoryError:
            logger.critical('Cannot open file "%s" as it is a directory', file)
            raise ChecklistLoadError
        except PermissionError:
            logger.critical('Cannot open file "%s" due to insufficient permissions', file)
            raise ChecklistLoadError

    def load_checklist(self, file):
        """Load a checklist with all of its pages, tasks and facts from a YAML file and process all imports."""

        logger.info('Loading checklist data from "%s"', file)

        try:
            return self.process_checklist(self.load_yaml(file), file.parent)
        except ChecklistLoadError:
            logger.critical('Loading checklist data from file failed')
            sys.exit(1)

    def save_checklist(self, file, checklist, overwrite=True):
        """Save a checklist and its pages, tasks and facts to a YAML file."""

        logger.info('Saving checklist data to "%s"', file.resolve())

        # Before writing the file, dump to a separate stream instead and check if we have an output.
        # This prevents saving an empty file if for some reason the dump fails.
        stream = io.StringIO()
        self.yaml.dump(checklist.to_dict(), stream)

        if stream.tell() == 0:
            logger.critical('Yaml dump failed and returned an empty result. File "%s" is left untouched', file)
            sys.exit(1)

        try:
            file.parent.mkdir(parents=True, exist_ok=True)
        except (NotADirectoryError, FileExistsError) as error:
            fallback_file = pathlib.Path.cwd() / file.name
            logger.error(
                'Cannot create parent directory "%s" (%s). Saving to "%s" instead',
                file.parent,
                error,
                fallback_file,
            )
            file = fallback_file

        if not overwrite:
            # Find a free file name
            original_file = file
            counter = 1

            while file.exists():
                file = original_file.with_name(f'{original_file.stem}_{counter}{original_file.suffix}')
                counter += 1

            if file != original_file:
                logger.warning('File "%s" already exists. Saving to "%s" instead', original_file, file)

        with open(file, mode='w' if overwrite else 'x', encoding='utf-8') as checklist_file:
            stream.seek(0)
            checklist_file.write(stream.read())

    def process_checklist(self, checklist, workdir):
        facts = {}

        if checklist is None:
            raise ValueError('Cannot load an empty checklist.')

        valid, message = utils.validate_dict_keys(
            checklist,
            {'title', 'pages'},
            {'description', 'report_path', 'version'},
            disallow_extra_keys=True,
        )

        if not valid:
            logger.critical('Failed to load checklist: %s', message)
            raise ChecklistLoadError

        page_list = checklist['pages']
        title = checklist['title']

        if not isinstance(title, str):
            logger.critical('Title field of checklist is not a string')
            raise ChecklistLoadError

        if not title or title.isspace():
            logger.warning('Title field of checklist is empty')

        if not isinstance(page_list, list):
            logger.critical('Checklist "%s" does not contain any pages', title)
            raise ChecklistLoadError

        if 'description' in checklist and not isinstance(checklist['description'], str):
            logger.critical('Description field of checklist "%s" is not a string', title)
            raise ChecklistLoadError

        if 'report_path' in checklist and not isinstance(checklist['report_path'], str):
            logger.critical('Report path field of checklist "%s" is not a string', title)
            raise ChecklistLoadError

        if 'version' in checklist and not isinstance(checklist['version'], str):
            logger.critical('Version field of checklist "%s" is not a string', title)
            raise ChecklistLoadError

        return models.Checklist(
            title,
            self.process_page_list(page_list, workdir, facts),
            facts,
            description=checklist.get('description'),
            report_path=checklist.get('report_path'),
            version=checklist.get('version'),
        )

    def process_page_list(self, page_list, workdir, facts):
        pages = []

        for page in page_list:
            if not isinstance(page, dict):
                logger.critical('Page data is not a mapping')
                raise ChecklistLoadError

            # Try to detect special directives before processing the page.
            try:
                page_directive, page_context = list(page.items())[0]
            except IndexError:
                page_directive = None
                page_context = None

            if page_directive == 'linuxfabrik.clf.import':
                if not isinstance(page_context, str):
                    logger.critical('Page import key is specified but its value is not a string')
                    raise ChecklistLoadError

                # Relative import paths should be relative to the checklist file.
                import_path = pathlib.Path(page_context)
                computed_import_path = import_path if import_path.is_absolute() else workdir / page_context

                logger.info('Importing pages from "%s"', computed_import_path)

                imported_page_list = self.load_yaml(computed_import_path)

                if not isinstance(imported_page_list, list):
                    logger.critical('Imported data is not a page list')
                    raise ChecklistLoadError

                pages.extend(self.process_page_list(imported_page_list, computed_import_path.parent, facts))
                continue

            pages.append(self.process_page(page, workdir, facts))

        return pages

    def process_page(self, page, workdir, facts):
        valid, message = utils.validate_dict_keys(page, {'title', 'tasks'}, optional_keys={'when'}, disallow_extra_keys=True)

        if not valid:
            logger.critical('Failed to load page: %s', message)
            raise ChecklistLoadError

        task_list = page['tasks']
        title = page['title']

        if not isinstance(title, str):
            logger.critical('Title field of page is not a string')
            raise ChecklistLoadError

        if not title or title.isspace():
            logger.critical('Title field of page is empty')
            raise ChecklistLoadError

        if not isinstance(task_list, list):
            logger.critical('Tasks field on page "%s" is not a list', title)
            raise ChecklistLoadError

        if task_list:
            tasks = self.process_task_list(page['tasks'], workdir, facts)
        else:
            logger.warning('Task list on page "%s" is empty', title)
            tasks = []

        return models.Page(title, tasks, page.get('when'))

    def process_task_list(self, task_list, workdir, facts):
        tasks = []

        for task in task_list:
            if not isinstance(task, dict):
                logger.critical('Task data is not a mapping')
                raise ChecklistLoadError

            task_items = list(task.items())

            if len(task_items) == 0:
                logger.critical('Task data is missing')
                raise ChecklistLoadError

            task_module, task_context = task_items[0]

            if task_module == 'linuxfabrik.clf.import':
                if not isinstance(task_context, str):
                    logger.critical('Task import key is specified but its value is not a string')
                    raise ChecklistLoadError

                # Relative import paths should be relative to the checklist file.
                import_path = pathlib.Path(task_context)
                computed_import_path = import_path if import_path.is_absolute() else workdir / task_context

                logger.info('Importing tasks from "%s"', computed_import_path)

                imported_task_list = self.load_yaml(computed_import_path)

                if not isinstance(imported_task_list, list):
                    logger.critical('Imported data is not a task list')
                    raise ChecklistLoadError

                tasks.extend(self.process_task_list(imported_task_list, computed_import_path.parent, facts))
                continue

            if not isinstance(task_context, dict):
                logger.critical('Task context is not a mapping')
                raise ChecklistLoadError

            fact_name = task.get('fact_name')
            value = task.get('value')
            when = task.get('when')

            unnamed_fact = None

            if fact_name is not None:
                if not fact_name.isidentifier():
                    logger.warning(
                        'Fact name "%s" is not a valid (Python) identifier. It cannot be referenced as a variable in Jinja templates',
                        fact_name,
                    )

                if fact_name.endswith('[]'):
                    logger.critical('Invalid fact name "%s". Fact names may not end with "[]"', fact_name)
                    # ChecklistFabrik uses `[]` as a suffix on HTML forms to differentiate lists from single values.
                    raise ChecklistLoadError

                if value is not None:
                    facts[fact_name] = value
            else:
                unnamed_fact = value

            tasks.append(models.Task(task_module, task_context, fact_name, when, unnamed_fact))

        return tasks
